import pandas as pd
from comet import download_model, load_from_checkpoint

from .LangDetectModel import FastTextLangDetectModel


class CometAggregationModel:
    model = None

    def __init__(
            self,
            model_id="Unbabel/wmt22-cometkiwi-da",
            get_lang_func=None,

    ):
        self.get_lang = get_lang_func
        if get_lang_func is None:
            fast_text_lang_detect_model = FastTextLangDetectModel()
            self.get_lang = fast_text_lang_detect_model.get_lang

        model_path = download_model(model_id)
        self.model = load_from_checkpoint(model_path)

    def get_blind_score(self, out_text_list, source_text, target_lang_code='en'):
        comet_data = [{
            "src": source_text[0],
            "mt": out_text_list[i],
            "ref": ""
        } for i in range(len(out_text_list))
        ]
        scores = self.model.predict(comet_data, batch_size=8, gpus=1)
        out_score = scores.scores
        for i in range(len(out_score)):
            if self.get_lang(out_text_list[i]) != target_lang_code:
                out_score[i] = 0
        return out_score

    def combine_preds(self, pred_dict, source_text, target_lang_code='en'):
        blind_results = {k: self.get_blind_score(pred_dict[k], source_text, target_lang_code=target_lang_code) for k in
                         pred_dict}
        score_df = pd.DataFrame.from_dict({k: blind_results[k] for k in blind_results}, orient="columns")
        best_pred_key = score_df.apply(lambda row: row.index[row.argmax()], axis=1).to_list()

        combined_pred = []
        for i in range(len(best_pred_key)):
            key = best_pred_key[i]
            combined_pred.append(pred_dict[key][i])

        return combined_pred



from .API_Model import APIModel


class OllamaModel(APIModel):
    def __init__(self, host='localhost', port=11434, endpoint='/api/generate', model="t_ragx_mistral",
                 protocol="http"):
        super().__init__(host=host, port=port, endpoint=endpoint, model=model,
                         protocol=protocol)



import logging
from operator import itemgetter
from typing import List, Union

import torch
from Levenshtein import distance
from elasticsearch import Elasticsearch
from elasticsearch import client as elastic_client
from tqdm.autonotebook import tqdm

from .BaseInputProcessor import BaseInputProcessor
from ..utils.heuristic import clean_text

logger = logging.getLogger("t_ragx")


def rerank_elastic_result(elastic_result, source_lang, search_term, top_k=5):
    if len(elastic_result) < 1:
        return []
    top_score = None
    result_list = []
    for r in elastic_result['hits']['hits']:
        if top_score is None:
            top_score = r['_score']
        if r['_score'] < top_score and len(result_list) > top_k:
            break

        if source_lang not in r['_source']:
            continue
        r['distance'] = distance(r['_source'][source_lang], search_term)
        result_list.append(r)

    result_list = sorted(result_list, key=itemgetter('distance'))
    return result_list[:top_k]


def search_single_elastic(es_client, index, search_term, source_lang, target_lang, top_k=10, request_timeout=50,
                          task_index=None, task_boost=1.2):
    index_list = [index]
    indices_boost = []
    if task_index is not None:
        index_list.append(task_index)
        indices_boost.append({
            task_index: task_boost
        })

    resp = es_client.search(
        index=index_list,
        body={
            "size": top_k,
            "indices_boost": indices_boost,
            "_source": {
                "includes": [source_lang, target_lang, 'source']
            },
            "query": {
                "bool": {
                    "must": [
                        {
                            "query_string": {
                                "query": search_term,
                                "fields": [
                                    source_lang
                                ],
                                "escape": True
                            }
                        },
                    ],
                    "filter": [
                        {"exists": {"field": target_lang}},
                    ]
                }
            }
        },
        request_timeout=request_timeout
    )
    return resp


def search_elastic_with_retry(es_client, index, search_term, source_lang, target_lang, top_k=10, retry=3,
                              task_index=None, task_boost=1.2):
    for i in range(retry):
        try:
            return search_single_elastic(es_client, index, search_term, source_lang, target_lang, top_k=top_k,
                                         task_index=task_index, task_boost=task_boost)
        except:
            pass
    logger.warning("elastic time out")
    return []


def batch_search_elastic(es_client, index, search_term_list, source_lang, target_lang, top_k=10, rerank_top_k=5,
                         pbar=False, task_index=None, task_boost=1.2, max_item_len=-1):
    bulk_result = []
    for search_term in tqdm(search_term_list, disable=(not pbar)):
        search_result = search_elastic_with_retry(es_client, index, search_term, source_lang, target_lang, top_k=top_k,
                                                  task_index=task_index, task_boost=task_boost)

        # truncate if the max_item_len variable is set
        for r in search_result['hits']['hits']:
            r['_source'][source_lang] = r['_source'][source_lang][:max_item_len]
            r['_source'][target_lang] = r['_source'][target_lang][:max_item_len]

        bulk_result.append(search_result)

    return [
        rerank_elastic_result(resp, source_lang, search_term, top_k=rerank_top_k) for search_term, resp in
        zip(search_term_list, bulk_result)
    ]


class ElasticInputProcessor(BaseInputProcessor):
    """
    The default input processor, rely on an elastic search database and pre-computed indexes
    """

    def __init__(self,
                 device=None,
                 ):
        self.device = device
        if device is None:
            self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
        super().__init__()

    def load_general_translation(self, elastic_index="translation_memory", elasticsearch_host: str = "localhost",
                                 elasticsearch_port: int = 9200, es_client: elastic_client = None, elastic_args={},
                                 elastic_client_args={}, **kwargs):
        """
        Load the general translation examples
        """

        # initiate the elastic index
        if es_client is None:
            es_client = Elasticsearch(
                elasticsearch_host,  # Elasticsearch endpoint
                port=elasticsearch_port,
                **elastic_client_args
            )

        self.es_client = es_client

        assert es_client.indices.exists(index=elastic_index)

        self.general_memory_elastic_index = elastic_index

        return

    def search_general_memory(self, *args, **kwargs):
        return self.search_memory(*args, **kwargs)

    def search_memory(self, text_list: Union[List[str], str], search_index: str = None, source_lang='ja',
                      target_lang='en', top_k=10,
                      rerank_top_k=None, max_item_len=500, pbar=False, task_index=None, task_boost=1.2,
                      **search_kwargs):
        """
        search general translation examples using elasticsearch
        """
        if isinstance(text_list, str):
            text_list = [text_list]

        text_list = [clean_text(t) for t in text_list]
        if search_index is None:
            search_index = self.general_memory_elastic_index

        if rerank_top_k is None:
            rerank_top_k = top_k

        search_result_list = batch_search_elastic(self.es_client, search_index, text_list, source_lang, target_lang,
                                                  top_k=top_k, rerank_top_k=rerank_top_k, pbar=pbar,
                                                  task_index=task_index, task_boost=task_boost,
                                                  max_item_len=max_item_len)

        processed_output = [[{'score': r['_score'], 'distance': r['distance']} | r['_source'] for r in search_result]
                            for search_result in search_result_list]

        # "normed_distance" is the Levenshtein
        for t, l in zip(text_list, processed_output):
            for d in l:
                d['normed_distance'] = d['distance'] / len(t)

        return processed_output



import abc

from transformers import AutoTokenizer, AutoModelForCausalLM

from .constants import LANG_BY_LANG_CODE


def pretext_to_text(pretext_list, max_sent=5):
    if pretext_list is None or len(pretext_list) < 1:
        return ""

    out_text = "Preceding text:\n"
    for source_text in pretext_list[:max_sent]:
        out_text += f"  {source_text}\n"
    return out_text + "\n"


def glossary_to_text(glossary):
    out_text = "Relevant Dictionary records:\n"
    for source_text in glossary:
        out_text += f"  {source_text}: {', '.join(glossary[source_text])}\n"
    return out_text


def trans_mem_to_text(trans_mem: list, source_lang_code='ja', target_lang_code='en'):
    if len(trans_mem) < 1:
        return ""
    out_text = "Examples translations:\n"
    count = 1
    for row in trans_mem:
        out_text += f""" {count}. \n   {row[source_lang_code]}\n   {row[target_lang_code]}\n"""
        count += 1
    return out_text


class BaseModel(metaclass=abc.ABCMeta):
    tokenizer = None
    model = None

    def __init__(self, model_id, adapter=None, tokenizer=None, model=None):

        if tokenizer is None:
            tokenizer = AutoTokenizer.from_pretrained(
                model_id,
                padding_side='left',
                truncation_side='left',
            )

            if tokenizer.pad_token_id is None:
                tokenizer.pad_token_id = tokenizer.unk_token_id
                tokenizer.pad_token = tokenizer.unk_token

        if model is None:
            model = AutoModelForCausalLM.from_pretrained(model_id, device_map="auto")
            if adapter is not None:
                if isinstance(adapter, list):
                    for a in adapter:
                        model.load_adapter(a)
                elif isinstance(adapter, str):
                    model.load_adapter(adapter)
                else:
                    ValueError("the adapter parameter must be either string or a list of strings")

            model = model.eval()

        self.model = model
        self.tokenizer = tokenizer

    def tokenize(self,
                 text_list=None,
                 tokenize_config=None
                 ):

        if text_list is None:
            text_list = []
        if tokenize_config is None:
            tokenize_config = {}

        default_tokenize_config = {
            'pad_to_multiple_of': 8,
            'padding': True,
            'truncation': True,
            'max_length': 2000,
            'return_tensors': 'pt',
            'add_special_tokens': False
        }
        for k in default_tokenize_config:
            if k not in tokenize_config:
                tokenize_config[k] = default_tokenize_config[k]

        return self.tokenizer.batch_encode_plus(text_list, **tokenize_config).to(self.model.device)

    def generate(self, tokenized_input, generation_config=None):
        if generation_config is None:
            generation_config = {
                'max_new_tokens': 100,
                'early_stopping': True,
                'eos_token_id': [self.tokenizer.eos_token_id],
                'pad_token_id': self.tokenizer.eos_token_id
            }

        for k in tokenized_input:
            tokenized_input[k] = tokenized_input[k].to(self.model.device)

        return self.model.generate(**tokenized_input, **generation_config)

    @staticmethod
    def clean_output(text):
        raise NotImplementedError

    def process_output(self, model_output, tokenized_input):
        translation_outputs = [
            o[len(i):]
            for o, i in zip(model_output.cpu().numpy(), tokenized_input['input_ids'].cpu().numpy())
        ]

        decoded_outputs = self.tokenizer.batch_decode(translation_outputs, skip_special_tokens=True)
        decoded_outputs = [
            self.clean_output(o) for o in decoded_outputs
        ]
        return decoded_outputs

    def batch_translate(self, batch_text: list,
                        source_lang_code="ja",
                        target_lang_code="en",
                        batch_search_result: list = None,
                        batch_pre_text: list = None,
                        tokenize_config=None,
                        generation_config=None
                        ):

        query_prompts = self.batch_build_prompt(
            text=batch_text,
            source_lang_code=source_lang_code,
            target_lang_code=target_lang_code,
            pre_text_list=batch_pre_text,
            search_result=batch_search_result
        )

        token_data = self.tokenize(query_prompts, tokenize_config)
        generation_output = self.generate(token_data, generation_config)
        translated_output = self.process_output(generation_output, token_data)
        return translated_output

    def translate(self, text: str,
                  source_lang_code="ja",
                  target_lang_code="en",
                  search_result: list = None,
                  pre_text: list = None,
                  tokenize_config=None,
                  generation_config=None
                  ):
        batch_text = [text]
        batch_pre_text = [pre_text]
        batch_search_result = [search_result]

        return self.batch_translate(
            batch_text,
            source_lang_code=source_lang_code,
            target_lang_code=target_lang_code,
            batch_search_result=batch_search_result,
            batch_pre_text=batch_pre_text,
            tokenize_config=tokenize_config,
            generation_config=generation_config
        )[0]

    def batch_build_prompt(self,
                           text: list,
                           source_lang_code="Japanese",
                           target_lang_code="English",
                           search_result: list = None,
                           pre_text_list: list = None,
                           ):

        if pre_text_list is not None:
            assert len(pre_text_list) == len(text)
        else:
            pre_text_list = [None] * len(text)

        if search_result is not None:
            assert len(search_result) == len(text)
        else:
            search_result = [None] * len(text)

        return [
            self.build_prompt(
                t,
                source_lang_code=source_lang_code,
                target_lang_code=target_lang_code,
                search_result=sr,
                pre_text=pt
            )
            for t, sr, pt in zip(text, search_result, pre_text_list)
        ]

    def build_prompt(self,
                     text,
                     source_lang_code="ja",
                     target_lang_code="en",
                     search_result=None,
                     pre_text: list = None
                     ):
        source_lang = LANG_BY_LANG_CODE[source_lang_code]
        target_lang = LANG_BY_LANG_CODE[target_lang_code]
        if search_result is None:
            search_result = {'glossary': [], 'memory': []}

        chat = [
            {"role": "user", "content": (
                "As a large language model, you are a trained expert in multiple languages. "
                "These are some references that might help you translating passages:\n"
                f"{glossary_to_text(search_result['glossary'])}{pretext_to_text(pre_text)}{trans_mem_to_text(search_result['memory'], source_lang_code=source_lang_code, target_lang_code=target_lang_code)}"
                f"Translate this {source_lang} passage to {target_lang} "
                "without additional questions, disclaimer, or explanations, but accurately and completely:"
                f"{text}"
            )},
        ]
        return self.tokenizer.apply_chat_template(chat, tokenize=False, )



from .InternLM2Model import InternLM2Model
from .LlamaCppPythonModel import LlamaCppPythonModel
from .MistralModel import MistralModel
from .OllamaModel import OllamaModel
from .OpenAIModel import OpenAIModel

"""Import all modules that exist in the current directory."""
# Ref https://stackoverflow.com/a/60861023/
from importlib import import_module
from pathlib import Path

__all__ = ["MistralModel", "InternLM2Model", "OllamaModel", "OpenAIModel", "LlamaCppPythonModel"]

for f in Path(__file__).parent.glob("*.py"):
    module_name = f.stem
    if (not module_name.startswith("_")) and (module_name not in globals()):
        import_module(f".{module_name}", __package__)
        __all__.append(module_name)
    del f, module_name
del import_module, Path



import requests

from .BaseModel import BaseModel
from ._utils import DummyTokenizer


class APIModel(BaseModel):
    tokenizer = DummyTokenizer()
    model = None

    def __init__(self, host='localhost', port='11434', endpoint='/api/generate', model="t_ragx_mistral",
                 protocol="http"):
        self.host = host
        self.port = port
        self.endpoint = endpoint
        self.protocol = protocol
        self.model = model
        super().__init__(model_id="Dummy", tokenizer=self.tokenizer, model=model)

    def generate(self, input_text_list, generation_config={}):
        if isinstance(input_text_list, str):
            input_text_list = [input_text_list]

        out_text = []
        for t in input_text_list:
            data = {
                "model": self.model,
                "prompt": t,
                'stream': False,
                **generation_config
            }

            r = requests.post(f"{self.protocol}://{self.host}:{self.port}{self.endpoint}", json=data)
            assert r.status_code == 200, f"Failed to generate {t}"
            out_text.append(r.json()['response'])

        return out_text

    def tokenize(self,
                 text_list=None,
                 *args, **kwargs
                 ):
        return text_list

    @staticmethod
    def clean_output(text):
        return text.strip()

    def process_output(self, model_output, tokenized_input):
        return [self.clean_output(t) for t in model_output]



import logging

from llama_cpp import Llama

from .BaseModel import BaseModel
from ._utils import DummyTokenizer

logging.getLogger("llama-cpp-python").setLevel(logging.WARNING)


class LlamaCppPythonModel(BaseModel):
    tokenizer = DummyTokenizer()
    model = None

    def __init__(self,
                 repo_id="rayliuca/TRagx-GGUF-NeuralOmniBeagle-7B",
                 filename="*Q4_K_M*",
                 # see https://huggingface.co/rayliuca/TRagx-GGUF-NeuralOmniBeagle-7B/tree/main
                 # for filename formats
                 model=None,
                 chat_format="mistral-instruct",
                 reset_model=True,
                 model_config={'n_ctx': 2048},
                 ):

        self.reset_model = reset_model
        if model is None:
            model = Llama.from_pretrained(
                repo_id=repo_id,
                filename=filename,
                chat_format=chat_format,
                **model_config
            )
        self.model = model
        super().__init__(model_id="Dummy", tokenizer=self.tokenizer, model=model)

    def generate(self, input_text_list, generation_config={}):
        if isinstance(input_text_list, str):
            input_text_list = [input_text_list]

        default_generation_config = {
            'max_tokens': 100,  # this is the max new token equivalent
            'stop': ["[INST]", "[/INST]", "<s>", "</s>"]
        }

        for k in default_generation_config:
            if k not in generation_config:
                generation_config[k] = default_generation_config[k]

        out_text = []
        for t in input_text_list:
            messages = [{
                "role": "user",
                "content": t
            }]

            if self.reset_model:
                self.model.reset()
            output = self.model.create_chat_completion(
                messages,
                **generation_config
            )

            out_text.append(output['choices'][0]['message']['content'])

        return out_text

    def tokenize(self,
                 text_list=None,
                 *args, **kwargs
                 ):
        return text_list

    @staticmethod
    def clean_output(text):
        return text.strip()

    def process_output(self, model_output, tokenized_input):
        return [self.clean_output(t) for t in model_output]



from typing import List, Union

import numpy as np
from tqdm.auto import tqdm

from .processors import ElasticInputProcessor
from .processors import BaseInputProcessor
from .models.AggregationModel import CometAggregationModel
from .models.BaseModel import BaseModel


class TRagx:
    """
    🦖🦖🦖
    Translation using LLM with Retrieval Augmented Generation (RAG)

    """
    aggregate_model = None

    def __init__(self,
                 generation_models: Union[BaseModel, List[BaseModel]],
                 aggregate_model=None,
                 input_processor: BaseInputProcessor = None,
                 ):
        """

        Args:
            generation_models: a list of the T-Ragx models
            aggregate_model: a model that would choose the best translation without references. Currently only support
                                t_ragx.models.AggregationModel.CometAggregationModel
            input_processor: a T-Ragx input processor
        """

        self.input_processor = input_processor
        if input_processor is None:
            self.input_processor = ElasticInputProcessor()

        self.generation_models = generation_models
        if not isinstance(generation_models, list):
            self.generation_models = [generation_models]

        if aggregate_model is not None:
            self.aggregate_model = aggregate_model
        else:
            if len(self.generation_models) > 1:
                self.aggregate_model = CometAggregationModel()

    def __call__(self, *args, **kwargs):
        return self.translate(*args, **kwargs)

    def translate(self, text, pre_text: list = None,
                  search_glossary=True, search_memory=True,
                  memory_search_args: dict = None,
                  glossary_search_args: dict = None,
                  prompt_args: List[dict] = None,
                  generation_args: List[dict] = None):
        pass

    def batch_translate(self,
                        text_list,
                        pre_text_list: list = None,
                        batch_size=1,
                        source_lang_code='ja',
                        target_lang_code='en',
                        search_glossary=True,
                        search_memory=True,
                        memory_search_args: dict = None,
                        glossary_search_args: dict = None,
                        tokenize_args: List[dict] = None,
                        prompt_args: List[dict] = None,
                        generation_args: List[dict] = None
                        ):

        if pre_text_list is None:
            pre_text_list = [None] * len(text_list)

        if memory_search_args is None:
            memory_search_args = {}
        if glossary_search_args is None:
            glossary_search_args = {}

        if prompt_args is None:
            prompt_args = [{}] * len(self.generation_models)

        if generation_args is None:
            generation_args = [{}] * len(self.generation_models)

        if tokenize_args is None:
            tokenize_args = [{}] * len(text_list)

        if generation_args is None:
            generation_args = [{}] * len(text_list)

        memory_results = [[]] * len(text_list)
        if search_memory:
            memory_results = self.input_processor.search_memory(text_list, **memory_search_args)

        glossary_results = [[]] * len(text_list)
        if search_glossary:
            glossary_results = self.input_processor.batch_search_glossary(text_list, **glossary_search_args)

        generation_output_dict = {}
        for model_idx, generation_model, p_args, tok_args, gen_args in zip(
                range(len(self.generation_models)),
                self.generation_models,
                prompt_args,
                tokenize_args,
                generation_args
        ):
            translated_text_list = []
            for batch_idx in tqdm(
                    np.array_split(list(range(len(text_list))), int(max(len(text_list) / batch_size, 1)))
            ):
                batch_text = [text_list[i] for i in batch_idx]
                batch_pre_text = [pre_text_list[i] for i in batch_idx]

                batch_search_result = [
                    {
                        'memory': memory_results[i],
                        'glossary': glossary_results[i],
                    }
                    for i in range(len(batch_idx))
                ]

                translated_text_list += generation_model.batch_translate(
                    batch_text,
                    source_lang_code=source_lang_code,
                    target_lang_code=target_lang_code,
                    batch_search_result=batch_search_result,
                    batch_pre_text=batch_pre_text,
                    tokenize_config=tok_args,
                    generation_config=gen_args
                )
            generation_output_dict[model_idx] = translated_text_list

        generation_output = generation_output_dict[0]
        if len(generation_output_dict) > 1:
            generation_output = self.aggregate_model.combine_preds(
                generation_output_dict, batch_text, target_lang_code=target_lang_code
            )
        return generation_output



import json
import urllib

import pandas as pd
from elasticsearch import Elasticsearch

from ..utils.elastic import upload_df

INDEX = "translation_memory_demo"
DEMO_MEMORY_FOLDER_URL = "https://l8u0.c18.e2-1.dev/t-ragx-public/memory/demo"

es_client = Elasticsearch()

data_index_url = f"{DEMO_MEMORY_FOLDER_URL}/index.json"
response = urllib.request.urlopen(data_index_url)
assert response.status == 200, "Cannot read the DEMO memory"

data_index = json.loads(response.read())

for data_file in data_index:
    file_url = f"{DEMO_MEMORY_FOLDER_URL}/{data_file['file_name']}"
    df = pd.read_parquet(file_url)
    upload_df(df, es_client, id_key=data_file['source_lang'])



import base64
import json
import os.path
import pathlib
import tempfile
import urllib.request
from hashlib import md5
import numpy as np

import requests

from ..utils.heuristic import clean_text


def serialize_str(s):
    return json.dumps(s, ensure_ascii=False)


def en_text_search(text, keyword):
    if len(keyword) > len(text):
        return False
    text = text.casefold()
    keyword = keyword.casefold()
    if f" {keyword} " in text:
        return True
    elif text == keyword:
        return True
    elif len(keyword) - 1 > len(text) and text[:len(keyword) + 1] == keyword + " ":
        return True
    elif len(keyword) - 1 > len(text) and text[-len(keyword) + 1:] == " " + keyword:
        return True
    return False


def merge_glossary_index(df):
    """
    merge the glossary records if the index is not unique
    otherwise df.to_dict("index") will throw error
    """

    dup_index = df[df.index.duplicated()].index
    for idx in dup_index:
        for c in df.columns:
            new_list = []
            for arr in df.loc[idx, c]:
                new_list += arr.tolist()
            df.loc[idx, c] = [np.array(list(set(new_list)))] * len(df.loc[idx, c])

    if len(dup_index) > 0:
        df = df[~df.index.duplicated(keep='first')]

    return df


# heuristic glossary retrieval
def get_glossary(text, glossary_index, max_k=10, lang_code='en', source_lang='ja'):
    text = clean_text(text)
    out_dict = {}
    count = 0
    for k in glossary_index:
        if lang_code not in glossary_index[k]:
            continue
        if (k in text and source_lang != 'en') or \
                (source_lang == 'en' and en_text_search(text, k)):
            skip_flag = False
            # check for glossary word being a component of a longer glossary word
            for ek in out_dict:
                if k.casefold() in ek.casefold():
                    skip_flag = True
                    break
            if skip_flag:
                continue

            out_dict[k] = glossary_index[k][lang_code].tolist()
            count += 1
            if count >= max_k:
                break

    return out_dict


def get_http_file_id(url):
    response = requests.head(url)
    # use ETag if available
    if 'ETag' in response.headers:
        return response.headers['ETag'].replace('"', "")

    # use encoded url path if ETag is not available
    return md5(base64.urlsafe_b64encode(url.encode())).hexdigest()


def file_cacher(file_path, tempfolder=None):
    """
    If the input file_path is a http url, cache the file (by ETag if possible) to local tempfolder

    Args:
        file_path:
        tempfolder:

    Returns:

    """
    if tempfolder is None:
        tempfolder = tempfile.gettempdir() + "/t_ragx"
        pathlib.Path(tempfolder).mkdir(parents=True, exist_ok=True)
    out_path = file_path
    if "http" in file_path:
        file_id = get_http_file_id(file_path)
        file_extension = pathlib.Path(file_path).suffix
        out_path = f"{tempfolder}/{file_id}{file_extension}"

        if not os.path.isfile(out_path):
            urllib.request.urlretrieve(file_path, out_path)

    return out_path



import abc
import typing

import datasets
import pandas as pd
import torch
from elasticsearch import Elasticsearch
from elasticsearch import client as elastic_client
from jinja2 import Template as JinjaTemplate
from tqdm.autonotebook import tqdm

from ._utils import get_glossary, file_cacher, merge_glossary_index
from .constants import DEFAULT_GLOSSARY_PARQUET_FOLDER
from ..utils.heuristic import clean_text


class BaseInputProcessor(metaclass=abc.ABCMeta):
    """
    The base input processor, use  huggingface datasets with elastic search for translation memories

    See the ElasticInputProcessor for using with Elasticsearch directly
    """

    def __init__(self,
                 device=None,
                 prompt_template: typing.Optional[typing.Union[str, JinjaTemplate]] = None
                 ):
        self.device = device
        if device is None:
            self.device = 'cuda' if torch.cuda.is_available() else 'cpu'

        self.es_client = None
        self.general_memory_elastic_index = None
        self.general_memory: typing.Optional[datasets.Dataset] = None
        self.task_memory: dict = {}

        self.general_glossary_dict: dict = {}
        self.task_glossary: dict = {}
        self.glossary_parquet_folder = DEFAULT_GLOSSARY_PARQUET_FOLDER

    def load_general_translation(self,
                                 parquet_path,
                                 index_key='ja',
                                 elasticsearch_host: str = "localhost",
                                 elasticsearch_port: int = 9200,
                                 es_client: elastic_client = None,
                                 dataset_args={},
                                 elastic_args={},
                                 elastic_client_args={}
                                 ):
        """
        Load the general translation examples
        """
        self.general_memory = datasets.Dataset.from_pandas(pd.read_parquet(parquet_path), **dataset_args)
        hasher = datasets.fingerprint.Hasher()
        memory_fingerprint = hasher.hash(self.general_memory)

        es_index_name = f"hf_{index_key}_{memory_fingerprint}"

        # initiate the elastic index
        if es_client is None:
            es_client = Elasticsearch(
                elasticsearch_host,  # Elasticsearch endpoint
                port=elasticsearch_port,
                **elastic_client_args
            )

        self.es_client = es_client

        if es_client.indices.exists(index=es_index_name):
            self.general_memory.load_elasticsearch_index(index_key, es_index_name=es_index_name)
        else:
            self.general_memory.add_elasticsearch_index(index_key, es_index_name=es_index_name,
                                                        es_client=es_client, **elastic_args)

        self.general_memory_elastic_index = index_key

        return

    def load_task_translation(self):
        """
        Load the general translation examples
        """
        raise NotImplementedError()
        pass

    def search_general_memory(self, text, search_index: str = None, k=4, max_item_len=500, **search_kwargs):
        """
        search general translation examples using elasticsearch
        """
        if search_index is None:
            search_index = self.general_memory_elastic_index

        mem_scores, mem_indices = self.general_memory.search_batch(search_index, text, k=k, **search_kwargs)

        ref_trans_data = [self.general_memory[midx] for midx in mem_indices]

        # truncate in case the example translations are too long
        processed_output = []
        for rtd, score_list in zip(ref_trans_data, mem_scores):
            wide_output = []
            key_list = list(rtd.keys())
            for i in range(len(rtd[key_list[0]])):
                wide_output.append({
                    k: rtd[k][i][:max_item_len] for k in key_list
                })
                wide_output[-1]['score'] = score_list[i]
            processed_output.append(wide_output)

        return processed_output

    def search_task_memory(self, client=None):
        """
        search in-task translation examples using elasticsearch
        """
        raise NotImplementedError()

    def load_general_glossary(self, glossary_parquet_folder=None, source_lang='ja', target_lang='en', encoding="utf8"):
        """
        Load the general glossary (i.e. wikidata title pair/ dictionary entries )
        format: {
            "original text" : ["translation 1", "translation 2"],
            "original text2" : ["translation 3", "translation 4"],
        }
        """
        if glossary_parquet_folder is not None:
            self.glossary_parquet_folder = glossary_parquet_folder
        self.general_glossary_dict[f"{source_lang}_{target_lang}"] = pd.read_parquet(
            file_cacher(f"{self.glossary_parquet_folder}/{source_lang}_{target_lang}.parquet")).to_dict("index")

        return

    def load_task_glossary(self, glossary_parquet_path, glossary_index):
        # raise NotImplementedError()
        task_glossary_df = pd.read_parquet(file_cacher(glossary_parquet_path))
        clean_index_dict = {k: clean_text(k) for k in task_glossary_df.index}
        task_glossary_df.rename(index=clean_index_dict, inplace=True)

        task_glossary_df = merge_glossary_index(task_glossary_df)

        glossary_dict = task_glossary_df.to_dict("index")

        self.task_glossary[glossary_index] = glossary_dict

        return

    def batch_search_glossary(self, text_list, max_k=10, task_index=None, search_general_glossary=True, max_workers=8,
                              chunksize=1, k=None, source_lang='ja', target_lang='en', pbar=False):
        def _temp_search_glossary(text, max_k=max_k, task_index=task_index,
                                  search_general_glossary=search_general_glossary):
            return self.search_glossary(text, max_k=max_k, task_index=task_index,
                                        search_general_glossary=search_general_glossary, source_lang=source_lang,
                                        target_lang=target_lang)
            pass

        return [_temp_search_glossary(t) for t in tqdm(text_list, disable=(not pbar))]

    def search_glossary(self, text, max_k=10, task_index=None, search_general_glossary=True, k=None, source_lang='ja',
                        target_lang='en'):
        if k is not None:
            max_k = k
        text = clean_text(text)

        found_glossary = {}

        if task_index is not None:
            found_glossary.update(self.search_task_glossary(text, task_index, max_k=max_k, target_lang=target_lang))

        if len(found_glossary) < max_k:
            general_glossary = self.search_general_glossary(text, max_k=max_k - len(found_glossary),
                                                            source_lang=source_lang, target_lang=target_lang)
            for k in general_glossary:
                skip_flag = False
                for existing_key in found_glossary:
                    if k in existing_key:
                        # ignore glossary words being a component of a longer glossary word
                        skip_flag = True
                        break
                if skip_flag:
                    continue

                if k in found_glossary:
                    found_glossary[k] = list(set(found_glossary[k] + general_glossary[k]))
                else:
                    found_glossary[k] = general_glossary[k]

        return found_glossary

    def search_general_glossary(self, text, max_k=10, source_lang='ja', target_lang='en'):
        if f"{source_lang}_{target_lang}" not in self.general_glossary_dict:
            self.load_general_glossary(self.glossary_parquet_folder, source_lang=source_lang, target_lang=target_lang)
        return get_glossary(clean_text(text), self.general_glossary_dict[f"{source_lang}_{target_lang}"], max_k=max_k,
                            lang_code=target_lang, source_lang=source_lang)
        # raise NotImplementedError()

    def search_task_glossary(self, text, task_index, max_k=10, source_lang='ja', target_lang='en'):
        return get_glossary(clean_text(text), self.task_glossary[task_index], max_k=max_k, lang_code=target_lang,
                            source_lang=source_lang)



import logging
import sys
from typing import Optional, Dict

from colorama import Fore, Back, Style


class ColoredFormatter(logging.Formatter):
    """Colored log formatter."""

    def __init__(self, *args, colors: Optional[Dict[str, str]] = None, **kwargs) -> None:
        """Initialize the formatter with specified format strings."""

        super().__init__(*args, **kwargs)

        self.colors = colors if colors else {}

    def format(self, record) -> str:
        """Format the specified record as text."""

        record.color = self.colors.get(record.levelname, '')
        record.reset = Style.RESET_ALL

        return super().format(record)


formatter = ColoredFormatter(
    '{asctime} {color}[{levelname:.1s}] {message}{reset}',
    style='{', datefmt='%Y-%m-%d %H:%M:%S',
    colors={
        'DEBUG': Fore.CYAN + Style.BRIGHT,
        'INFO': Fore.GREEN + Style.BRIGHT,
        'WARNING': Fore.YELLOW + Style.BRIGHT,
        'ERROR': Fore.RED + Style.BRIGHT,
        'CRITICAL': Fore.RED + Style.BRIGHT + Back.YELLOW,
    }
)

handler = logging.StreamHandler(sys.stdout)
handler.setFormatter(formatter)

logging.basicConfig()
logger = logging.getLogger("t_ragx")

logger.handlers[:] = []
logger.addHandler(handler)



from setuptools import find_packages, setup

setup(
    name='t_ragx',
    package_dir={"": "src"},
    packages=find_packages("src"),
    version='0.0.7',
    url="https://github.com/rayliuca/T-Ragx",
    description='Enhancing translation with RAG-powered large language models',
    long_description=open('README.md', encoding="utf8").read(),
    long_description_content_type='text/markdown',
    author='Ray Liu',
    email='ray@rayliu.ca',
    license='MIT',
    project_urls={
        "Bug Reports": "https://github.com/rayliuca/T-Ragx/issues",
        "Source": "https://github.com/rayliuca/T-Ragx",
    },

    install_requires=[
        'transformers>=4.38',
        'elasticsearch',
        'OpenCC',
        'levenshtein',
        'unbabel-comet>=2.2.1',
    ],
    keywords=[
        'machine-translation',
        'computer-aided-translation',
        'computer-assisted-translation',
        'rag',
        'retrieval-augmented-generation',
        'large-language-models',
        'llm',
    ]
)



import re

from .BaseModel import BaseModel


class InternLM2Model(BaseModel):
    tokenizer = None
    model = None

    def __init__(self, model_id="rayliuca/TRagx-Merged-internlm2-7b", adapter=None, tokenizer=None, model=None):
        super().__init__(model_id=model_id, adapter=adapter, tokenizer=tokenizer, model=model)

    def tokenize(self,
                 text_list=None,
                 tokenize_config=None
                 ):

        if text_list is None:
            text_list = []
        if tokenize_config is None:
            tokenize_config = {}

        default_tokenize_config = {
            'pad_to_multiple_of': 8,
            'padding': True,
            'truncation': True,
            'max_length': 2000,
            'return_tensors': 'pt',
            'add_special_tokens': False
        }
        for k in default_tokenize_config:
            if k not in tokenize_config:
                tokenize_config[k] = default_tokenize_config[k]

        return self.tokenizer.batch_encode_plus(text_list, **tokenize_config).to(self.model.device)

    @staticmethod
    def clean_output(text):
        special_tok_q = " ?/[//INST/] ?"
        return re.sub(special_tok_q, "", text.replace("//", "")).strip()

    def process_output(self, model_output, tokenized_input):
        translation_outputs = [
            o[len(i):]
            for o, i in zip(model_output.cpu().numpy(), tokenized_input['input_ids'].cpu().numpy())
        ]

        decoded_outputs = self.tokenizer.batch_decode(translation_outputs, skip_special_tokens=True)
        decoded_outputs = [
            self.clean_output(o) for o in decoded_outputs
        ]
        return decoded_outputs



import json
import logging
from hashlib import sha1

import numpy as np
import pandas as pd
from elasticsearch import Elasticsearch
from tqdm.notebook import tqdm

from .heuristic import clean_text, is_noise
from .heuristic import lang_detect as heuristic_lang_detect
from ..processors.constants import DEFAULT_MEMORY_INDEX
from ..models.constants import LANG_BY_LANG_CODE

logger = logging.getLogger("t_ragx")


def index_doc(df, index="translation_memory_demo"):
    """
    Formatted index action generator helper to help upload records to Elasticsearch

    Args:
        df:
        index:

    Returns:

    """
    for record in df.to_dict(orient="records"):
        # pop none
        for k in record:
            if record[k] is None:
                record.pop(k)
        yield ('{ "index" : { "_index" : "%s", "_id": "%s"}}' % (
            index, sha1(record[record['id_key']].encode('utf8')).hexdigest()))
        yield json.dumps(record, default=int)


def upsert_doc(df: pd.DataFrame, index: str = None):
    """
    Formatted upsert action generator helper to help upload records to Elasticsearch

    Args:
        df:
        index:

    Returns:

    """
    if index is None:
        index = DEFAULT_MEMORY_INDEX

    for record in df.to_dict(orient="records"):
        # pop none
        pop_list = []
        for k in record:
            if record[k] is None:
                pop_list.append(k)

        for k in pop_list:
            record.pop(k)
        yield ('{ "update" : {"_index" : "%s", "_id" : "%s", "retry_on_conflict" : 3} }' % (
            index, sha1(record[record['id_key']].encode('utf8')).hexdigest()))
        yield '{ "doc" : %s, "doc_as_upsert" : true }' % json.dumps(record, default=int)


def filter_df(df: pd.DataFrame, source_lang: str = 'ja', lang_cols: list = None):
    if lang_cols is None:
        lang_cols = list(LANG_BY_LANG_CODE.keys())

    lang_cols = list(set(lang_cols).intersection(df.columns))

    df.dropna(subset=lang_cols, how='all', inplace=True)
    df.drop_duplicates(subset=[source_lang], inplace=True)
    df[source_lang] = df[source_lang].apply(clean_text)
    df = df[~df[source_lang].map(is_noise)]
    df.reset_index(drop=True, inplace=True)

    for c in lang_cols:
        df = df[~df[c].str.contains("\n", na=False)]

    for c in lang_cols:
        if c in ['ja', 'zh']:
            str_len = df[c].str.len()
            df = df[
                ((350 > str_len) & (str_len > 4)) | (str_len.isna())
                ]
        elif c in ['en']:
            word_count = df[c].str.split(" ").str.len()
            df = df[
                ((100 > word_count) & (word_count > 3)) | (word_count.isna())
                ]

    for c in lang_cols:
        detected_lang = df[c].apply(heuristic_lang_detect)
        df = df[(c == detected_lang) | (detected_lang.isna())]

    df.reset_index(drop=True, inplace=True)

    return df


def upload_df(df: pd.DataFrame, es_client: Elasticsearch(), id_key: str = 'ja', batch_size: int = 10000,
              index: str = None) -> None:
    """
    upload_df

    Args:
        df:
        es_client:
        id_key: The language column to hash (sha1) as ID. Duplicate records with common id will be merged.
                        id_key should be in df.columns
        batch_size:
        index: Defaulted to be "translation_memory". Should be explicitly set for in-task memories

    Returns:

    """
    df = filter_df(df, source_lang=id_key)
    df['id_key'] = id_key
    if len(df) < 1:
        print("Empty dataset")
        return
    batch_idx = np.array_split(range(len(df)), max(int(len(df) / batch_size), 1))
    for select_idx in tqdm(batch_idx):
        try:
            r = es_client.bulk(upsert_doc(df.iloc[select_idx]), index)  # return a dict
        except:
            raise r


def csv_to_elastic(file_path,
                   id_key='ja',
                   elasticsearch_host: str = "localhost",
                   elasticsearch_port: int = 9200,
                   es_client: Elasticsearch() = None,
                   batch_size=10000,
                   read_csv_config: dict = {},
                   index=None,
                   elastic_client_args: dict = {}):
    """
    Upload a CSV file to Elasticsearch
    The input csv should be parallel texts with the language code as their header
    For example:
        | ja  | en        | zh    |
        |-----|-----------|-------|
        | 例1 | example 1 | 範例1 |
        |     |           |       |
        |     |           |       |


    Args:

        file_path:
        id_key: The language column to hash (sha1) as ID. Duplicate records with common id will be merged.
                        id_key should be in df.columns
        elasticsearch_host:
        elasticsearch_port:
        es_client:
        batch_size:
        read_csv_config:
        index: Defaulted to be "translation_memory". Should be explicitly set for in-task memories
        elastic_client_args:

    Returns:

    """

    if es_client is None:
        es_client = Elasticsearch(
            elasticsearch_host,  # Elasticsearch endpoint
            port=elasticsearch_port,
            **elastic_client_args
        )

    df = pd.read_csv(file_path, **read_csv_config)
    assert len(df.columns) > 1, "The CSV file has only one column"

    if len(set(df.columns).intersection(LANG_BY_LANG_CODE.keys())) < 2:
        logger.warning(f"The columns of the CSV are {df.columns}")

    upload_df(df, es_client, id_key=id_key, batch_size=batch_size, index=index)



from openai import OpenAI

from .API_Model import APIModel
from ._utils import DummyTokenizer as BaseDummyTokenizer


class DummyTokenizer(BaseDummyTokenizer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def apply_chat_template(self, conversation, *args, **kwargs):
        return conversation


class OpenAIModel(APIModel):
    tokenizer = DummyTokenizer()
    model = None

    def __init__(self, host='localhost', port=11434, endpoint='/v1', model="t_ragx_mistral",
                 protocol="http", api_key='ollama'):
        super().__init__(host=host, port=port, endpoint=endpoint, model=model,
                         protocol=protocol)

        self.openai_client = OpenAI(
            base_url=f"{self.protocol}://{self.host}:{self.port}{self.endpoint}",
            api_key=api_key,
        )

    def generate(self, input_chat_list, generation_config={}):
        out_text = []
        for chat in input_chat_list:
            chat_completion = self.openai_client.chat.completions.create(
                messages=chat,
                model=self.model,
                **generation_config
            )
            out_text.append(chat_completion.choices[0].message.content.strip())

        return out_text



import abc
from typing import Union

import fasttext
import regex
from huggingface_hub import hf_hub_download


class BaseLangDetectModel(metaclass=abc.ABCMeta):
    """
    Heuristic lang detect model by counting the occurrence of Japanese and Chinese characters
    Only supports ja, zh, en outputs
    """

    def __init__(
            self,
            *arg,
            **kwargs
    ):
        pass

    @staticmethod
    def lang_detect(text, filter=False, filter_thrush=0.5) -> Union[str, None]:
        if text is None:
            return None
        ja_text = r"[\p{Katakana}\p{Hiragana}]"
        en_text = r"[a-zA-Z]"
        zh_text = r"\p{han}"

        text_len_dict = {
            'ja': len(regex.findall(ja_text, text)),
            'en': len(regex.findall(en_text, text)),
            'zh': len(regex.findall(zh_text, text))
        }

        max_key = max(text_len_dict, key=text_len_dict.get)

        if filter:
            total_len = len(text)
            if text_len_dict[max_key] / total_len < filter_thrush:
                return None
        return max_key


class FastTextLangDetectModel(BaseLangDetectModel):
    """
    Detect the language model via the Facebook language identification model
    """

    def __init__(self, repo_id="facebook/fasttext-language-identification", filename="model.bin", hf_hub_args={}, *arg,
                 **kwargs):
        super().__init__(*arg, **kwargs)
        fasttext_langid_model_path = hf_hub_download(repo_id=repo_id,
                                                     filename=filename, **hf_hub_args)
        self.model = fasttext.load_model(fasttext_langid_model_path)

    def get_lang(self, text):
        for lang in self.model.predict(text.replace("\n", " "), k=-1)[0]:
            if "_Han" in lang or "_Tibt" in lang:
                return "zh"
            elif "_Latn" in lang:
                return "en"
            elif "_Jpan" in lang:
                return "ja"
        return self.model.predict(text, k=-1)[0].split("__")[-1]



class DummyTokenizer:
    eos_token_id = None
    eos_token = None

    pad_token_id = None
    pad_token = None

    unk_token_id = None
    unk_token = None

    def __init__(self, *args, **kwargs):
        pass

    def batch_decode(self, *args, **kwargs):
        raise NotImplementedError

    def batch_encode_plus(self, *args, **kwargs):
        raise NotImplementedError

    def apply_chat_template(self, conversation, *args, **kwargs):
        return conversation[-1]['content']



import re

import regex
import unicodedata


def lang_detect(text):
    if text is None:
        return None

    ja_text = r"[\p{Katakana}\p{Hiragana}]"
    en_text = r"[a-zA-Z]"
    zh_text = r"\p{han}"

    text_len_dict = {
        'ja': len(regex.findall(ja_text, text)),
        'en': len(regex.findall(en_text, text)),
        'zh': len(regex.findall(zh_text, text))
    }

    max_key = max(text_len_dict, key=text_len_dict.get)

    # # filtering?
    # total_len = len(text)
    # if text_len_dict[max_key] / total_len < 0.5:
    #     return None
    return max_key


def clean_text(text):
    return unicodedata.normalize('NFKD', text).strip()


def is_date(text):
    # date_format_list = ['%Y年%m月%d', '%Y年%m月%d日', '%Y年', '%m月%d日', '%d日', '%m月']
    date_format_list = ['\d{1,}年\d{1,}月', '\d{1,}年\d{1,}月\d{1,}日', '\d{1,}年', '\d{1,}月\d{1,}日', '\d{1,}日',
                        '\d{1,}月']
    for date_format in date_format_list:
        # if not pd.isna(pd.to_datetime(text, format=date_format, errors='coerce')):
        if re.match('^' + date_format + '$', text) is not None:
            return True
    return False


def is_number(text):
    return re.match("^\d{1,}\.?\d*$", text) is not None


# assert is_number("54564.564") == True
# assert is_number("24778") == True
# assert is_number("24.7.78") == False
# assert is_number("24..778") == False
# assert is_number("1a24778") == False
# assert is_number("a124778") == False
# assert is_number("a124778f") == False


def is_noise(text):
    """
    many text in wiki or web crawled data is just number or date, which is not very useful as memory or glossary
    """
    if is_number(text):
        return True

    if is_date(text):
        return True
    return False
